"""
The well-known Trap problem.

The Trap function is similar to OneMax, but swaps the worst possible solution
with the global optimum. This means that the best-possible objective value, 0,
is reached for the string of all `False` bits. The worst objective value, `n`,
is reached by the strings with exactly one `True` bit.

1. Stefan Droste, Thomas Jansen, and Ingo Wegener. On the Analysis of the
   (1+1) Evolutionary Algorithm. *Theoretical Computer Science.*
   276(1-2):51-81. April 2002.
   doi: https://doi.org/10.1016/S0304-3975(01)00182-7
2. Siegfried Nijssen and Thomas Bäck. An Analysis of the Behavior of
   Simplified Evolutionary Algorithms on Trap Functions. IEEE Transactions on
   Evolutionary Computation. 7(1):11-22. 2003.
   doi: https://doi.org/10.1109/TEVC.2002.806169
3. Thomas Weise, Zhize Wu, Xinlu Li, and Yan Chen. Frequency Fitness
   Assignment: Making Optimization Algorithms Invariant under Bijective
   Transformations of the Objective Function Value. *IEEE Transactions on
   Evolutionary Computation* 25(2):307-319. April 2021. Preprint available at
   arXiv:2001.01416v5 [cs.NE] 15 Oct 2020.
   https://dx.doi.org/10.1109/TEVC.2020.3032090
4. Thomas Weise, Zhize Wu, Xinlu Li, Yan Chen, and Jörg Lässig. Frequency
   Fitness Assignment: Optimization without Bias for Good Solutions can be
   Efficient. *IEEE Transactions on Evolutionary Computation (TEVC)*.
   27(4):980-992. August 2023.
   doi: https://doi.org/10.1109/TEVC.2022.3191698
"""
from typing import Callable, Final, Iterator, cast

import numba  # type: ignore
import numpy as np

from moptipy.examples.bitstrings.bitstring_problem import BitStringProblem


@numba.njit(nogil=True, cache=True)
def trap(x: np.ndarray) -> int:
    """
    Compute the trap objective value.

    :param x: the np array
    :return: the trap function value

    >>> print(trap(np.array([True, True, False, False, False])))
    4
    >>> print(trap(np.array([True, False, True, False, False])))
    4
    >>> print(trap(np.array([False, True,  True, False, False])))
    4
    >>> print(trap(np.array([True, True, True, True, True])))
    1
    >>> print(trap(np.array([False, True, True, True, True])))
    2
    >>> print(trap(np.array([False, False, False, False, False])))
    0
    >>> print(trap(np.array([False, True,  False, False, False])))
    5
    >>> print(trap(np.array([False, True,  True, True, False])))
    3

    # n = 1 and 0 true bits
    >>> trap(np.array([0]))
    0

    # n = 1 and 1 true bit
    >>> trap(np.array([1]))
    1

    # n = 2 and 0 true bits
    >>> trap(np.array([0, 0]))
    0

    # n = 2 and 1 true bit
    >>> trap(np.array([0, 1]))
    2

    # n = 2 and 1 true bit
    >>> trap(np.array([0, 1]))
    2

    # n = 2 and 1 true bit
    >>> trap(np.array([0, 1]))
    2

    # n = 2 and 2 true bits
    >>> trap(np.array([1, 1]))
    1

    # n = 3 and 0 true bits
    >>> trap(np.array([0, 0, 0]))
    0

    # n = 3 and 1 true bit
    >>> trap(np.array([0, 1, 0]))
    3

    # n = 3 and 1 true bit
    >>> trap(np.array([1, 0, 0]))
    3

    # n = 3 and 1 true bit
    >>> trap(np.array([0, 0, 1]))
    3

    # n = 3 and 2 true bits
    >>> trap(np.array([0, 1, 1]))
    2

    # n = 3 and 2 true bits
    >>> trap(np.array([1, 0, 1]))
    2

    # n = 3 and 2 true bits
    >>> trap(np.array([0, 1, 1]))
    2

    # n = 3 and 3 true bits
    >>> trap(np.array([1, 1, 1]))
    1

    # n = 4 and 0 true bits
    >>> trap(np.array([0, 0, 0, 0]))
    0

    # n = 4 and 1 true bit
    >>> trap(np.array([1, 0, 0, 0]))
    4

    # n = 4 and 1 true bit
    >>> trap(np.array([1, 0, 0, 0]))
    4

    # n = 4 and 1 true bit
    >>> trap(np.array([0, 0, 0, 1]))
    4

    # n = 4 and 2 true bits
    >>> trap(np.array([1, 1, 0, 0]))
    3

    # n = 4 and 2 true bits
    >>> trap(np.array([0, 0, 1, 1]))
    3

    # n = 4 and 2 true bits
    >>> trap(np.array([0, 1, 1, 0]))
    3

    # n = 4 and 3 true bits
    >>> trap(np.array([0, 1, 1, 1]))
    2

    # n = 4 and 3 true bits
    >>> trap(np.array([1, 1, 1, 0]))
    2

    # n = 4 and 3 true bits
    >>> trap(np.array([1, 1, 0, 1]))
    2

    # n = 4 and 4 true bits
    >>> trap(np.array([1, 1, 1, 1]))
    1

    # n = 5 and 0 true bits
    >>> trap(np.array([0, 0, 0, 0, 0]))
    0

    # n = 5 and 1 true bit
    >>> trap(np.array([0, 0, 0, 0, 1]))
    5

    # n = 5 and 1 true bit
    >>> trap(np.array([0, 0, 0, 1, 0]))
    5

    # n = 5 and 1 true bit
    >>> trap(np.array([1, 0, 0, 0, 0]))
    5

    # n = 5 and 2 true bits
    >>> trap(np.array([1, 0, 1, 0, 0]))
    4

    # n = 5 and 2 true bits
    >>> trap(np.array([1, 0, 1, 0, 0]))
    4

    # n = 5 and 2 true bits
    >>> trap(np.array([0, 1, 0, 0, 1]))
    4

    # n = 5 and 3 true bits
    >>> trap(np.array([1, 0, 1, 1, 0]))
    3

    # n = 5 and 3 true bits
    >>> trap(np.array([0, 1, 1, 0, 1]))
    3

    # n = 5 and 3 true bits
    >>> trap(np.array([0, 1, 1, 0, 1]))
    3

    # n = 5 and 4 true bits
    >>> trap(np.array([1, 0, 1, 1, 1]))
    2

    # n = 5 and 4 true bits
    >>> trap(np.array([1, 1, 1, 0, 1]))
    2

    # n = 5 and 4 true bits
    >>> trap(np.array([1, 1, 0, 1, 1]))
    2

    # n = 5 and 5 true bits
    >>> trap(np.array([1, 1, 1, 1, 1]))
    1

    # n = 6 and 0 true bits
    >>> trap(np.array([0, 0, 0, 0, 0, 0]))
    0

    # n = 6 and 1 true bit
    >>> trap(np.array([0, 0, 0, 0, 1, 0]))
    6

    # n = 6 and 1 true bit
    >>> trap(np.array([0, 0, 0, 0, 0, 1]))
    6

    # n = 6 and 1 true bit
    >>> trap(np.array([0, 0, 1, 0, 0, 0]))
    6

    # n = 6 and 2 true bits
    >>> trap(np.array([1, 0, 0, 1, 0, 0]))
    5

    # n = 6 and 2 true bits
    >>> trap(np.array([1, 0, 0, 0, 1, 0]))
    5

    # n = 6 and 2 true bits
    >>> trap(np.array([1, 0, 0, 0, 1, 0]))
    5

    # n = 6 and 3 true bits
    >>> trap(np.array([0, 1, 1, 1, 0, 0]))
    4

    # n = 6 and 3 true bits
    >>> trap(np.array([1, 1, 0, 0, 0, 1]))
    4

    # n = 6 and 3 true bits
    >>> trap(np.array([0, 1, 1, 1, 0, 0]))
    4

    # n = 6 and 4 true bits
    >>> trap(np.array([1, 0, 0, 1, 1, 1]))
    3

    # n = 6 and 4 true bits
    >>> trap(np.array([1, 0, 0, 1, 1, 1]))
    3

    # n = 6 and 4 true bits
    >>> trap(np.array([0, 1, 1, 1, 1, 0]))
    3

    # n = 6 and 5 true bits
    >>> trap(np.array([0, 1, 1, 1, 1, 1]))
    2

    # n = 6 and 5 true bits
    >>> trap(np.array([1, 1, 1, 1, 0, 1]))
    2

    # n = 6 and 5 true bits
    >>> trap(np.array([1, 1, 1, 1, 1, 0]))
    2

    # n = 6 and 6 true bits
    >>> trap(np.array([1, 1, 1, 1, 1, 1]))
    1

    # n = 7 and 0 true bits
    >>> trap(np.array([0, 0, 0, 0, 0, 0, 0]))
    0

    # n = 7 and 1 true bit
    >>> trap(np.array([0, 0, 1, 0, 0, 0, 0]))
    7

    # n = 7 and 1 true bit
    >>> trap(np.array([0, 0, 0, 0, 0, 1, 0]))
    7

    # n = 7 and 1 true bit
    >>> trap(np.array([0, 0, 0, 1, 0, 0, 0]))
    7

    # n = 7 and 2 true bits
    >>> trap(np.array([0, 0, 1, 0, 0, 0, 1]))
    6

    # n = 7 and 2 true bits
    >>> trap(np.array([0, 0, 1, 0, 0, 1, 0]))
    6

    # n = 7 and 2 true bits
    >>> trap(np.array([1, 0, 1, 0, 0, 0, 0]))
    6

    # n = 7 and 3 true bits
    >>> trap(np.array([0, 0, 1, 1, 0, 1, 0]))
    5

    # n = 7 and 3 true bits
    >>> trap(np.array([1, 0, 0, 1, 1, 0, 0]))
    5

    # n = 7 and 3 true bits
    >>> trap(np.array([1, 1, 0, 0, 0, 0, 1]))
    5

    # n = 7 and 4 true bits
    >>> trap(np.array([0, 1, 0, 0, 1, 1, 1]))
    4

    # n = 7 and 4 true bits
    >>> trap(np.array([1, 1, 0, 0, 1, 1, 0]))
    4

    # n = 7 and 4 true bits
    >>> trap(np.array([1, 0, 1, 0, 0, 1, 1]))
    4

    # n = 7 and 5 true bits
    >>> trap(np.array([1, 1, 0, 1, 1, 1, 0]))
    3

    # n = 7 and 5 true bits
    >>> trap(np.array([1, 1, 1, 1, 1, 0, 0]))
    3

    # n = 7 and 5 true bits
    >>> trap(np.array([1, 1, 0, 1, 0, 1, 1]))
    3

    # n = 7 and 6 true bits
    >>> trap(np.array([1, 1, 1, 1, 1, 1, 0]))
    2

    # n = 7 and 6 true bits
    >>> trap(np.array([1, 1, 0, 1, 1, 1, 1]))
    2

    # n = 7 and 6 true bits
    >>> trap(np.array([1, 1, 1, 1, 1, 1, 0]))
    2

    # n = 7 and 7 true bits
    >>> trap(np.array([1, 1, 1, 1, 1, 1, 1]))
    1

    # n = 8 and 0 true bits
    >>> trap(np.array([0, 0, 0, 0, 0, 0, 0, 0]))
    0

    # n = 8 and 1 true bit
    >>> trap(np.array([0, 0, 0, 0, 1, 0, 0, 0]))
    8

    # n = 8 and 1 true bit
    >>> trap(np.array([0, 1, 0, 0, 0, 0, 0, 0]))
    8

    # n = 8 and 1 true bit
    >>> trap(np.array([0, 0, 1, 0, 0, 0, 0, 0]))
    8

    # n = 8 and 2 true bits
    >>> trap(np.array([1, 0, 0, 0, 0, 1, 0, 0]))
    7

    # n = 8 and 2 true bits
    >>> trap(np.array([0, 1, 0, 0, 1, 0, 0, 0]))
    7

    # n = 8 and 2 true bits
    >>> trap(np.array([0, 1, 0, 0, 1, 0, 0, 0]))
    7

    # n = 8 and 3 true bits
    >>> trap(np.array([0, 1, 0, 0, 0, 1, 1, 0]))
    6

    # n = 8 and 3 true bits
    >>> trap(np.array([0, 1, 1, 0, 0, 1, 0, 0]))
    6

    # n = 8 and 3 true bits
    >>> trap(np.array([1, 1, 0, 0, 1, 0, 0, 0]))
    6

    # n = 8 and 4 true bits
    >>> trap(np.array([1, 0, 1, 0, 1, 0, 0, 1]))
    5

    # n = 8 and 4 true bits
    >>> trap(np.array([0, 1, 1, 1, 1, 0, 0, 0]))
    5

    # n = 8 and 4 true bits
    >>> trap(np.array([1, 1, 1, 0, 0, 1, 0, 0]))
    5

    # n = 8 and 5 true bits
    >>> trap(np.array([1, 1, 0, 0, 0, 1, 1, 1]))
    4

    # n = 8 and 5 true bits
    >>> trap(np.array([1, 1, 1, 0, 0, 1, 1, 0]))
    4

    # n = 8 and 5 true bits
    >>> trap(np.array([0, 0, 1, 1, 0, 1, 1, 1]))
    4

    # n = 8 and 6 true bits
    >>> trap(np.array([0, 1, 1, 1, 1, 1, 1, 0]))
    3

    # n = 8 and 6 true bits
    >>> trap(np.array([1, 1, 0, 1, 1, 1, 0, 1]))
    3

    # n = 8 and 6 true bits
    >>> trap(np.array([1, 1, 1, 0, 0, 1, 1, 1]))
    3

    # n = 8 and 7 true bits
    >>> trap(np.array([1, 1, 1, 1, 1, 1, 0, 1]))
    2

    # n = 8 and 7 true bits
    >>> trap(np.array([1, 1, 1, 1, 1, 1, 1, 0]))
    2

    # n = 8 and 7 true bits
    >>> trap(np.array([1, 0, 1, 1, 1, 1, 1, 1]))
    2

    # n = 8 and 8 true bits
    >>> trap(np.array([1, 1, 1, 1, 1, 1, 1, 1]))
    1

    # n = 9 and 0 true bits
    >>> trap(np.array([0, 0, 0, 0, 0, 0, 0, 0, 0]))
    0

    # n = 9 and 1 true bit
    >>> trap(np.array([0, 0, 0, 0, 0, 1, 0, 0, 0]))
    9

    # n = 9 and 1 true bit
    >>> trap(np.array([0, 0, 0, 0, 0, 0, 0, 1, 0]))
    9

    # n = 9 and 1 true bit
    >>> trap(np.array([0, 1, 0, 0, 0, 0, 0, 0, 0]))
    9

    # n = 9 and 2 true bits
    >>> trap(np.array([1, 0, 0, 0, 1, 0, 0, 0, 0]))
    8

    # n = 9 and 2 true bits
    >>> trap(np.array([0, 0, 0, 0, 0, 0, 0, 1, 1]))
    8

    # n = 9 and 2 true bits
    >>> trap(np.array([0, 0, 0, 1, 0, 0, 0, 1, 0]))
    8

    # n = 9 and 3 true bits
    >>> trap(np.array([0, 1, 0, 0, 0, 1, 0, 0, 1]))
    7

    # n = 9 and 3 true bits
    >>> trap(np.array([1, 0, 0, 1, 1, 0, 0, 0, 0]))
    7

    # n = 9 and 3 true bits
    >>> trap(np.array([1, 1, 0, 0, 0, 0, 1, 0, 0]))
    7

    # n = 9 and 4 true bits
    >>> trap(np.array([1, 0, 1, 1, 0, 0, 1, 0, 0]))
    6

    # n = 9 and 4 true bits
    >>> trap(np.array([1, 0, 1, 0, 0, 0, 1, 1, 0]))
    6

    # n = 9 and 4 true bits
    >>> trap(np.array([1, 0, 0, 1, 0, 0, 1, 1, 0]))
    6

    # n = 9 and 5 true bits
    >>> trap(np.array([1, 0, 1, 1, 0, 0, 0, 1, 1]))
    5

    # n = 9 and 5 true bits
    >>> trap(np.array([0, 1, 0, 1, 0, 1, 1, 0, 1]))
    5

    # n = 9 and 5 true bits
    >>> trap(np.array([1, 0, 0, 0, 1, 0, 1, 1, 1]))
    5

    # n = 9 and 6 true bits
    >>> trap(np.array([1, 1, 1, 0, 0, 1, 0, 1, 1]))
    4

    # n = 9 and 6 true bits
    >>> trap(np.array([0, 1, 1, 0, 1, 1, 0, 1, 1]))
    4

    # n = 9 and 6 true bits
    >>> trap(np.array([1, 0, 1, 0, 1, 1, 1, 0, 1]))
    4

    # n = 9 and 7 true bits
    >>> trap(np.array([1, 1, 1, 1, 1, 1, 1, 0, 0]))
    3

    # n = 9 and 7 true bits
    >>> trap(np.array([1, 0, 1, 1, 1, 1, 1, 0, 1]))
    3

    # n = 9 and 7 true bits
    >>> trap(np.array([1, 1, 0, 1, 1, 1, 1, 0, 1]))
    3

    # n = 9 and 8 true bits
    >>> trap(np.array([0, 1, 1, 1, 1, 1, 1, 1, 1]))
    2

    # n = 9 and 8 true bits
    >>> trap(np.array([1, 1, 1, 1, 1, 1, 1, 1, 0]))
    2

    # n = 9 and 8 true bits
    >>> trap(np.array([1, 1, 1, 1, 1, 0, 1, 1, 1]))
    2

    # n = 9 and 9 true bits
    >>> trap(np.array([1, 1, 1, 1, 1, 1, 1, 1, 1]))
    1

    # n = 10 and 0 true bits
    >>> trap(np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0]))
    0

    # n = 10 and 1 true bit
    >>> trap(np.array([0, 0, 0, 0, 0, 1, 0, 0, 0, 0]))
    10

    # n = 10 and 1 true bit
    >>> trap(np.array([0, 0, 1, 0, 0, 0, 0, 0, 0, 0]))
    10

    # n = 10 and 1 true bit
    >>> trap(np.array([0, 0, 0, 0, 1, 0, 0, 0, 0, 0]))
    10

    # n = 10 and 2 true bits
    >>> trap(np.array([1, 0, 0, 0, 0, 1, 0, 0, 0, 0]))
    9

    # n = 10 and 2 true bits
    >>> trap(np.array([0, 1, 0, 0, 0, 0, 0, 0, 0, 1]))
    9

    # n = 10 and 2 true bits
    >>> trap(np.array([0, 0, 0, 1, 0, 1, 0, 0, 0, 0]))
    9

    # n = 10 and 3 true bits
    >>> trap(np.array([0, 1, 0, 0, 0, 0, 0, 1, 0, 1]))
    8

    # n = 10 and 3 true bits
    >>> trap(np.array([0, 1, 0, 0, 0, 1, 0, 0, 0, 1]))
    8

    # n = 10 and 3 true bits
    >>> trap(np.array([1, 0, 0, 1, 0, 0, 0, 1, 0, 0]))
    8

    # n = 10 and 4 true bits
    >>> trap(np.array([0, 0, 1, 0, 1, 0, 0, 1, 1, 0]))
    7

    # n = 10 and 4 true bits
    >>> trap(np.array([1, 0, 0, 0, 0, 0, 1, 0, 1, 1]))
    7

    # n = 10 and 4 true bits
    >>> trap(np.array([1, 0, 0, 1, 0, 1, 0, 0, 1, 0]))
    7

    # n = 10 and 5 true bits
    >>> trap(np.array([1, 1, 0, 1, 0, 1, 0, 0, 0, 1]))
    6

    # n = 10 and 5 true bits
    >>> trap(np.array([1, 0, 1, 0, 1, 0, 0, 0, 1, 1]))
    6

    # n = 10 and 5 true bits
    >>> trap(np.array([0, 1, 1, 1, 0, 1, 0, 0, 1, 0]))
    6

    # n = 10 and 6 true bits
    >>> trap(np.array([0, 1, 1, 0, 0, 1, 1, 1, 0, 1]))
    5

    # n = 10 and 6 true bits
    >>> trap(np.array([0, 0, 0, 1, 1, 0, 1, 1, 1, 1]))
    5

    # n = 10 and 6 true bits
    >>> trap(np.array([1, 1, 0, 1, 0, 1, 1, 0, 0, 1]))
    5

    # n = 10 and 7 true bits
    >>> trap(np.array([0, 1, 1, 0, 1, 1, 1, 1, 0, 1]))
    4

    # n = 10 and 7 true bits
    >>> trap(np.array([0, 1, 1, 1, 1, 1, 1, 0, 1, 0]))
    4

    # n = 10 and 7 true bits
    >>> trap(np.array([1, 1, 1, 1, 1, 0, 1, 1, 0, 0]))
    4

    # n = 10 and 8 true bits
    >>> trap(np.array([1, 1, 0, 0, 1, 1, 1, 1, 1, 1]))
    3

    # n = 10 and 8 true bits
    >>> trap(np.array([1, 0, 1, 1, 1, 0, 1, 1, 1, 1]))
    3

    # n = 10 and 8 true bits
    >>> trap(np.array([1, 1, 1, 1, 0, 0, 1, 1, 1, 1]))
    3

    # n = 10 and 9 true bits
    >>> trap(np.array([1, 1, 1, 1, 1, 1, 0, 1, 1, 1]))
    2

    # n = 10 and 9 true bits
    >>> trap(np.array([1, 1, 1, 1, 0, 1, 1, 1, 1, 1]))
    2

    # n = 10 and 9 true bits
    >>> trap(np.array([1, 1, 1, 1, 1, 1, 1, 1, 0, 1]))
    2

    # n = 10 and 10 true bits
    >>> trap(np.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1]))
    1
    """
    n: Final[int] = len(x)
    res: Final[int] = int(x.sum())
    return 0 if res <= 0 else n - res + 1


class Trap(BitStringProblem):
    """The trap problem."""

    def __init__(self, n: int) -> None:
        """
        Initialize the trap objective function.

        :param n: the dimension of the problem

        >>> print(Trap(55).n)
        55
        >>> print(Trap(4).evaluate(np.array([True, True, False, True])))
        2
        """
        super().__init__(n)
        self.evaluate = trap  # type: ignore

    def __str__(self) -> str:
        """
        Get the name of the trap objective function.

        :return: `trap_` + length of string

        >>> print(Trap(33))
        trap_33
        """
        return f"trap_{self.n}"

    @classmethod
    def default_instances(
            cls: type, scale_min: int = 3, scale_max: int = 333) \
            -> Iterator[Callable[[], "Trap"]]:
        """
        Get the 77 default instances of the :class:`Trap` problem.

        :param scale_min: the minimum permitted scale, by default `3`
        :param scale_max: the maximum permitted scale, by default `333`
        :returns: a sequence of default :class:`Trap` instances

        >>> len(list(Trap.default_instances()))
        77

        >>> [x() for x in Trap.default_instances()]
        [trap_3, trap_4, trap_5, trap_6, trap_7, trap_8, trap_9, trap_10, \
trap_11, trap_12, trap_13, trap_14, trap_15, trap_16, trap_17, trap_18, \
trap_19, trap_20, trap_21, trap_22, trap_23, trap_24, trap_25, trap_26, \
trap_27, trap_28, trap_29, trap_30, trap_31, trap_32, trap_33, trap_36, \
trap_40, trap_41, trap_42, trap_44, trap_48, trap_49, trap_50, trap_55, \
trap_59, trap_60, trap_64, trap_66, trap_70, trap_77, trap_79, trap_80, \
trap_81, trap_85, trap_88, trap_90, trap_96, trap_99, trap_100, trap_107, \
trap_111, trap_121, trap_125, trap_128, trap_144, trap_149, trap_169, \
trap_170, trap_192, trap_196, trap_199, trap_200, trap_222, trap_225, \
trap_243, trap_256, trap_269, trap_289, trap_300, trap_324, trap_333]
        """
        return cast("Iterator[Callable[[], Trap]]",
                    super().default_instances(  # type: ignore
                        scale_min, scale_max))



"""
A Hybrid SA-FEA Algorithm: the `SAFEA-N`.

This hybrid algorithm has the following features:

- The new solution of the FEA strand is always copied to the SA strand.
- The new solution of the SA strand is copied over to the FEA strand if it is
  better than the current solution of the FEA strand.
- The H-table is updated by both strands.
- The FEA strand always toggles back to the SA strand.
- The SA strand toggles to the FEA strand if it did not improve for a certain,
  increasing `time limit`.
- Every time the SA strand toggles over to the FEA strand, `time_limit` is
  incremented by 1.
"""
from collections import Counter
from math import exp
from typing import Callable, Final

from numpy.random import Generator
from pycommons.types import type_error

from moptipy.algorithms.modules.temperature_schedule import TemperatureSchedule
from moptipy.algorithms.so.ffa.ffa_h import create_h, log_h
from moptipy.api.algorithm import Algorithm1
from moptipy.api.operators import Op0, Op1
from moptipy.api.process import Process
from moptipy.utils.logger import KeyValueLogSection


class SAFEAN(Algorithm1):
    """An implementation of the SAFEA-N."""

    def __init__(self, op0: Op0, op1: Op1, schedule: TemperatureSchedule,
                 log_h_tbl: bool = False) -> None:
        """
        Create the SAFEA-N.

        :param op0: the nullary search operator
        :param op1: the unary search operator
        :param schedule: the temperature schedule to use
        :param log_h_tbl: should we log the H table?
        """
        if not isinstance(schedule, TemperatureSchedule):
            raise type_error(schedule, "schedule", TemperatureSchedule)
        if not isinstance(log_h_tbl, bool):
            raise type_error(log_h_tbl, "log_h_tbl", bool)
        super().__init__(f"safeaN_{schedule}", op0, op1)
        #: True if we should log the H table, False otherwise
        self.__log_h_tbl: Final[bool] = log_h_tbl
        #: the temperature schedule
        self.schedule: Final[TemperatureSchedule] = schedule

    def solve(self, process: Process) -> None:
        """
        Apply the SAFEA-N to an optimization problem.

        :param process: the black-box process object
        """
        # Create records for old and new point in the search space.
        x_ea = process.create()  # record for current solution of the SA
        x_fea = process.create()  # record for current solution of the FEA
        x_new = process.create()  # record for new solution

        # Obtain the random number generator.
        random: Final[Generator] = process.get_random()

        # Put function references in variables to save time.
        evaluate: Final[Callable] = process.evaluate  # the objective
        should_terminate: Final[Callable] = process.should_terminate
        temperature: Final[Callable[[int], float]] = self.schedule.temperature
        r01: Final[Callable[[], float]] = random.random  # random from [0, 1)
        xcopy: Final[Callable] = process.copy  # copy(dest, source)
        op0: Final[Callable] = self.op0.op0  # the nullary operator
        op1: Final[Callable] = self.op1.op1  # the unary operator

        h, ofs = create_h(process)  # Allocate the h-table

        # Start at a random point in the search space and evaluate it.
        op0(random, x_ea)  # Create 1 solution randomly and
        y_ea: int | float = evaluate(x_ea) + ofs  # evaluate it.
        xcopy(x_fea, x_ea)  # FEA and SA start with the same initial solution.
        y_fea: int | float = y_ea

        sa_max_no_lt_moves: int = 1  # maximum no-improvement moves for SA
        sa_no_lt_moves: int = 0  # current no-improvement moves
        use_ffa: bool = False  # We start with the SA branch.
        tau: int = 0  # The iteration index, needs to be 0 at first cmp.

        while not should_terminate():  # Until we need to quit...
            # Sample and evaluate new solution.
            op1(random, x_new, x_fea if use_ffa else x_ea)
            y_new: int | float = evaluate(x_new) + ofs
            h[y_new] += 1  # type: ignore  # Always update H.

            if use_ffa:  # The FEA branch uses FFA.
                use_ffa = False  # Always toggle use from FFA to SA.
                sa_no_lt_moves = 0  # Reset the SA no-improv move counter.

                h[y_fea] += 1  # type: ignore  # Update H for FEA solution.
                if h[y_new] <= h[y_fea]:  # type: ignore  # FEA acceptance.
                    xcopy(x_ea, x_new)  # Copy solution also to SA.
                    x_fea, x_new = x_new, x_fea
                    y_fea = y_new
                else:  # FEA does not accept, but we always copy to the SA,
                    x_ea, x_new = x_new, x_ea  # so we quickly swap here.
                y_ea = y_new  # we always copy the solution over to the SA

            else:  # SA branch performs simulated annealing.
                h[y_ea] += 1  # type: ignore  # Update H in *both* branches.

                if y_new < y_fea:  # Is new solution better than the FEA one?
                    xcopy(x_fea, x_new)  # Copy solution over to FEA.
                    y_fea = y_new  # And store the objective value.

                if (y_new <= y_ea) or (  # Accept if <= or if SA criterion
                        r01() < exp((y_ea - y_new) / temperature(tau))):
                    x_ea, x_new = x_new, x_ea  # Accept new solution.
                    y_new, y_ea = y_ea, y_new  # Swap values (for line below).
                    if y_new > y_ea:  # Check if we did an actual improvement.
                        sa_no_lt_moves = 0  # non-improving moves counter = 0.
                        continue  # We can jump to the next iteration.

                sa_no_lt_moves += 1  # Increase non-improved counter.
                if sa_no_lt_moves >= sa_max_no_lt_moves:  # Toggle: SA to FEA.
                    sa_max_no_lt_moves += 1  # Increment limit by one.
                    use_ffa = True  # Toggle to FFA.
            tau += 1  # Step the iteration index.

        if not self.__log_h_tbl:
            return  # we are done here

        # After we are done, we want to print the H-table.
        if h[y_ea] == 0:  # type: ignore  # Fix the H-table for the case
            h = Counter()   # that only one FE was performed: In this case,
            h[y_ea] = 1  # make Counter with only a single 1 value inside.

        log_h(process, h, ofs)  # log the H-table

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log all parameters of the SAFEA-N algorithm.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        with logger.scope("ts") as ts:
            self.schedule.log_parameters_to(ts)



"""
A nullary operator filling an integer list with random values.

>>> from moptipy.utils.nputils import rand_generator
>>> gen = rand_generator(12)

>>> space = IntSpace(4, -2, 2)
>>> op = Op0Random(space)
>>> xx = space.create()

>>> op.op0(gen, xx)
>>> xx
array([-1,  2,  2,  1], dtype=int8)
>>> op.op0(gen, xx)
>>> xx
array([-2, -2, -1, -1], dtype=int8)
>>> op.op0(gen, xx)
>>> xx
array([ 1,  0, -2,  2], dtype=int8)

>>> space = IntSpace(5, 12000, 20000)
>>> op = Op0Random(space)
>>> xx = space.create()

>>> op.op0(gen, xx)
>>> xx
array([15207, 19574, 18014, 12515, 14406], dtype=int16)
>>> op.op0(gen, xx)
>>> xx
array([16080, 13596, 13434, 14148, 16655], dtype=int16)
"""


from typing import Final

import numpy as np
from numpy.random import Generator
from pycommons.types import type_error

from moptipy.api.operators import Op0
from moptipy.spaces.intspace import IntSpace


class Op0Random(Op0):
    """Fill a bit string with random values."""

    def __init__(self, space: IntSpace) -> None:
        """
        Create the Nullary Integer-Space operator.

        :param space: the integer space
        """
        if not isinstance(space, IntSpace):
            raise type_error(space, "space", IntSpace)
        #: the inclusive lower bound
        self.__lb: Final[int] = space.min_value
        #: the exclusive upper bound
        self.__ub: Final[int] = space.max_value + 1

    def op0(self, random: Generator, dest: np.ndarray) -> None:
        """
        Fill the string `dest` with random values.

        :param random: the random number generator
        :param dest: the bit string to fill. Afterward, its contents will
            be random.
        """
        np.copyto(dest, random.integers(
            self.__lb, self.__ub, dest.shape, dest.dtype))

    def __str__(self) -> str:
        """
        Get the name of this operator.

        :return: "uniform"
        """
        return "uniform"
